unit DrBobFTP;
{============================================================================}
{ DrBobFTP is Copyright (c) 1999 by Bob Swart (aka Dr.Bob - www.drbob42.com) }
{                                                                            }
{ TBFTP wraps around WININET and provides the File Transfer Protocol         }
{ Made available as freeware contribution for Delphi programmers world-wide. }
{============================================================================}
{
  example usage:

  program FTP;
  uses
    SysUtils, DrBobFTP;
  begin
    with TBFTP.Create(nil) do
    try
      Server := 'ftp.server.com';
      UserID := 'bob';
      Password := '********';
      Connect;
      ChangeDir('ftp');
      GetFile('headconv.zip','headconv.zip')
    finally
      Free // disconnect
    end
  end.
}
{$H+}
interface
uses
  Windows, Classes, SysUtils;

Type
  HInternet = Pointer;

  TBoolNotifyEvent = procedure (Sender: TObject; Success: Boolean) of Object;

  TBftp = class(TComponent)
  private
    FContext: Integer;
    FINet: HInternet;
    FFTPHandle: HInternet;
    FUserID: ShortString;
    FPassword: ShortString;
    FServer: ShortString;
    FDirectory: ShortString;
    FSubDirs: TStringList;
    FFiles: TStringList;
    FConfirmDelete: Boolean;
  private
    FOnNewDir: TNotifyEvent;
    FOnGetFile: TBoolNotifyEvent;
    FOnSendFile: TBoolNotifyEvent;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Connect: Boolean;
    function FindFiles: TStringList;
    function ChangeDir(SubDir: ShortString): Boolean;
    function GoBackDir: Boolean;
    function CreateDirectory(SubDir: ShortString): Boolean;
    function RemoveDirectory(SubDir: ShortString): Boolean;
    function GetFile(FTPFile, NewFile: ShortString): Boolean;
    function SendFile(FTPFile, NewFile: ShortString): Boolean;
    function DeleteFile(FTPFile: ShortString): Boolean;
  private
  { Private field declarations }
    FAbout, Dummy: ShortString;
  published
    property About: ShortString read FAbout write Dummy;
    property UserID: ShortString read FUserID write FUserID;
    property Password: ShortString read FPassword write FPassword;
    property Server: ShortString read FServer write FServer;
    property Directory: ShortString read FDirectory;
    property SubDirs: TStringList read FSubDirs;
    property Files: TStringList read FFiles;
    property ConfirmDelete: Boolean read FConfirmDelete write FConfirmDelete;
  published
    property OnNewDir: TNotifyEvent read FOnNewDir write FOnNewDir;
    property OnGetFile: TBoolNotifyEvent read FOnGetFile write FOnGetFile;
    property OnSendFile: TBoolNotifyEvent read FOnSendFile write FOnSendFile;
  end {TBftp};

implementation
uses
  Forms, Dialogs, Controls;

  { WININET.PAS }

const
  winetdll = 'wininet.dll';
  INTERNET_SERVICE_FTP = 1;
  FTP_TRANSFER_TYPE_BINARY = $00000002;

function InternetOpen(lpszAgent: PChar; dwAccessType: DWORD;
  lpszProxy, lpszProxyBypass: PChar; dwFlags: DWORD): HINTERNET; stdcall;
  external winetdll name 'InternetOpenA';

function InternetCloseHandle(hInet: HINTERNET): BOOL; stdcall;
  external winetdll name 'InternetCloseHandle';

function InternetConnect(hInet: HINTERNET; lpszServerName: PChar;
  nServerPort: Word; lpszUsername: PChar; lpszPassword: PChar;
  dwService: DWORD; dwFlags: DWORD; dwContext: DWORD): HINTERNET; stdcall;
  external winetdll name 'InternetConnectA';

function FtpGetCurrentDirectory(hConnect: HINTERNET;
  lpszCurrentDirectory: PChar; var lpdwCurrentDirectory: DWORD): BOOL; stdcall;
  external winetdll name 'FtpGetCurrentDirectoryA';

function FtpFindFirstFile(hConnect: HINTERNET; lpszSearchFile: PChar;
  var lpFindFileData: TWin32FindData; dwFlags: DWORD;
  dwContext: DWORD): HINTERNET; stdcall;
  external winetdll name 'FtpFindFirstFileA';

function InternetFindNextFile(hFind: HINTERNET; lpvFindData: Pointer): BOOL; stdcall;
  external winetdll name 'InternetFindNextFileA';

function FtpSetCurrentDirectory(hConnect: HINTERNET; lpszDirectory: PChar): BOOL; stdcall;
  external winetdll name 'FtpSetCurrentDirectoryA';

function FtpCreateDirectory(hConnect: HINTERNET; lpszDirectory: PChar): BOOL; stdcall;
  external winetdll name 'FtpCreateDirectoryA';

function FtpRemoveDirectory(hConnect: HINTERNET; lpszDirectory: PChar): BOOL; stdcall;
  external winetdll name 'FtpRemoveDirectoryA';

function FtpGetFile(hConnect: HINTERNET; lpszRemoteFile: PChar;
  lpszNewFile: PChar; fFailIfExists: BOOL; dwFlagsAndAttributes: DWORD;
  dwFlags: DWORD; dwContext: DWORD): BOOL stdcall;
  external winetdll name 'FtpGetFileA';

function FtpPutFile(hConnect: HINTERNET; lpszLocalFile: PChar;
  lpszNewRemoteFile: PChar; dwFlags: DWORD; dwContext: DWORD): BOOL; stdcall;
  external winetdll name 'FtpPutFileA';

function InternetGetLastResponseInfo(var lpdwError: DWORD; lpszBuffer: PChar;
  var lpdwBufferLength: DWORD): BOOL; stdcall;
  external winetdll name 'InternetGetLastResponseInfoA';

function FtpDeleteFile(hConnect: HINTERNET; lpszFileName: PChar): BOOL; stdcall;
  external winetdll name 'FtpDeleteFileA';

  { FTP }

constructor TBftp.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FAbout := 'TBFTP 5.0 (c) 1999 by Bob Swart (aka Dr.Bob - www.drbob42.com)';
  FSubDirs := TStringList.Create;
  FFiles := TStringList.Create;
  FINet := InternetOpen('DrBob42', 0, nil, nil, 0);
  FConfirmDelete := False
end {Create};

destructor TBftp.Destroy;
begin
  if FINet <> nil then InternetCloseHandle(FINet);
  if FFtpHandle <> nil then InternetCloseHandle(FFtpHandle);
  FSubDirs.Free;
  FSubDirs := nil;
  FFiles.Free;
  FFIles := nil;
  inherited Destroy
end {Destroy};

function TBftp.Connect: Boolean;
var
  Len: DWORD;
begin
  FContext := 255;
  FServer[Length(FServer)+1] := #0;
  FUserID[Length(FUserID)+1] := #0;
  FPassword[Length(FPassword)+1] := #0;
  FftpHandle := InternetConnect(FINet, @FServer[1], 0,
                                @FUserID[1], @FPassWord[1],
                                Internet_Service_Ftp, 0, FContext);
  if FFtpHandle = nil then
    Result := False
  else
  begin
    Len := 0;
    ftpGetCurrentDirectory(FFTPHandle, @FDirectory[1], Len); { get Len }
    FDirectory[0] := Chr(Len);
    ftpGetCurrentDirectory(FFTPHandle, @FDirectory[1], Len); { get Dir }
    FindFiles;
    Result := True
  end;
  Application.ProcessMessages
end {Connect};

function TBftp.FindFiles: TStringList;
var
  FindData: TWin32FindData;
  FindHandle: HInternet;
begin
  FindHandle := FtpFindFirstFile(FFTPHandle, '*.*', FindData, 0, 0);
  FFiles.Clear;
  FSubDirs.Clear;
  FSubDirs.Add('[..]');
  if FindHandle <> nil then
  begin
    repeat
      if FindData.dwFileAttributes = FILE_ATTRIBUTE_DIRECTORY then
        FSubDirs.Add(Format('[%s]', [FindData.CFilename]))
      else FFiles.Add(Format('%s', [FindData.CFilename{FindData.nFileSizeLow div 1024}]))
    until not InternetFindnextFile(FindHandle, @FindData);
    InternetCloseHandle(FindHandle)
  end;
  Application.ProcessMessages;
  Result := FFiles
end {FindFiles};

function TBftp.GoBackDir: Boolean;
var
  S: ShortString;
  i: Integer;
begin
  S := FDirectory;
  i := Length(S);
  repeat
    Dec(i)
  until (i = 1) or (S[i] = '/');
  Delete(S,i,1+Length(S)-i);
  if S <> '' then
  begin
    ChangeDir(S);
    Result := True
  end
  else
  begin
    ChangeDir('/');
    Result := False
  end
end {GoBackDir};

function TBftp.ChangeDir(SubDir: ShortString): Boolean;
var
  Len: DWORD;
begin
  if SubDir <> '' then
  begin
    SubDir[Length(SubDir)+1] := #0;
    FtpSetCurrentDirectory(FFTPHandle, @SubDir[1])
  end;
  FindFiles;
  Len := 0;
  ftpGetCurrentDirectory(FFTPHandle, @FDirectory[1], Len); { get Len }
  FDirectory[0] := Chr(Len);
  ftpGetCurrentDirectory(FFTPHandle, @FDirectory[1], Len); { get Dir }
  if FDirectory[Len] <> '/' then FDirectory := FDirectory + '/';
  if Assigned(FOnNewDir) then FOnNewDir(Self);
  Result := True
end {ChangeDir};

function TBftp.CreateDirectory(SubDir: ShortString): Boolean;
begin
  SubDir[Length(SubDir)+1] := #0;
  Result := FtpCreateDirectory(FFTPHandle, @SubDir[1])
end {CreateDirectory};

function TBftp.RemoveDirectory(SubDir: ShortString): Boolean;
begin
  SubDir[Length(SubDir)+1] := #0;
  Result := FtpRemoveDirectory(FFTPHandle, @SubDir[1])
end {RemoveDirectory};

function TBftp.GetFile(FTPFile, NewFile: ShortString): Boolean;
begin
  FTPFile[Length(FTPFile)+1] := #0;
  NewFile[Length(NewFile)+1] := #0;
  Result := FtpGetFile(FFTPHandle, @FTPFile[1], @NewFile[1],
                       False, File_Attribute_Normal,
                       Ftp_Transfer_Type_Binary, 0);
  Application.ProcessMessages;
  if Assigned(FOnGetFile) then FOnGetFile(Self,Result)
end {GetFile};

function TBftp.SendFile(FTPFile, NewFile: ShortString): Boolean;
const
  Size: DWord = 3000;
var
  Error: DWord;
  S: String;
begin
  FTPFile[Length(FTPFile)+1] := #0;
  NewFile[Length(NewFile)+1] := #0;
  Result := FtpPutFile(FFTPHandle, @FTPFile[1], @NewFile[1],
                       Ftp_Transfer_Type_Binary, 0);
  if not Result then
  begin
    Error := GetLastError;
    ShowMessage(Format('Error Number: %d. Hex: %x', [Error, Error]));
    SetLength(S, Size);
    if not InternetGetLastResponseInfo(Error, PChar(S), Size) then
    begin
      Error := GetLastError;
      ShowMessage(Format('Error Number: %d. Hex: %x', [Error, Error]))
    end
    else
      ShowMessage(Format('Error Number: %d. Hex: %x Info: %s', [Error, Error, S]))
  end;
  Application.ProcessMessages;
  if Assigned(FOnGetFile) then FOnGetFile(Self,Result)
end {SendFile};

function TBftp.DeleteFile(FTPFile: ShortString): Boolean;
begin
  FTPFile[Length(FTPFile)+1] := #0;
  Result := not FConfirmDelete;
  if not Result then
    Result := MessageDlg('Delete '+FTPFile+'?', mtConfirmation, [mbYes, mbNo], 0) = mrYes;
  if Result then
    Result := FtpDeleteFile(FFTPHandle, @FTPFile[1]);
  Application.ProcessMessages
end {DeleteFile};

end.
